@@ -475,3 +475,78 @@
   ;; And we're done.
   (inst mov eax-tn 1) ;; exception-continue-search
   (inst ret))
+
+#+sb-assembling
+(define-assembly-routine (code-header-set (:return-style :none)) ()
+  (inst push eax-tn)
+  (inst push edx-tn)
+  (inst push edi-tn)
+  ;; stack: spill[3], ret-pc, object, index, value-to-store
+
+  (symbol-macrolet ((object (make-ea :dword :base esp-tn :disp 16))
+                    (word-index (make-ea :dword :base esp-tn :disp 20))
+                    (newval (make-ea :dword :base esp-tn :disp 24))
+                    (prefix #!+(and sb-thread (not win32)) :fs
+                            #!-(and sb-thread (not win32)) nil))
+    (flet ((thread-slot-ea (slot-index)
+             (make-ea :dword
+                      #!+(or (not sb-thread) win32) :base #!+(or (not sb-thread) win32) edi-tn
+                      :disp (ash slot-index word-shift))))
+      #!-sb-thread
+      (progn
+        ;; Load 'all_threads' into EDI (which was already spilled)
+        ;; as the register with which to access thread slots.
+        #!+sb-dynamic-core
+        (progn
+          (inst mov edi-tn
+                (make-ea :dword :disp (make-fixup "all_threads" :foreign-dataref)))
+          (inst mov edi-tn (make-ea :dword :base edi-tn)))
+        #!-sb-dynamic-core
+        (inst mov edi-tn (make-ea :dword :disp (make-fixup "all_threads" :foreign))))
+      #!+(and win32 sb-thread)
+      (inst mov edi-tn (make-ea :dword :disp +win32-tib-arbitrary-field-offset+) :fs)
+
+      (inst mov eax-tn object) ; object
+      (inst sub eax-tn (thread-slot-ea thread-dynspace-addr-slot) prefix)
+      (inst shr eax-tn (1- (integer-length gencgc-card-bytes)))
+      (pseudo-atomic ()
+        (assemble ()
+          (inst cmp eax-tn (thread-slot-ea thread-dynspace-card-count-slot)
+                prefix)
+          (inst jmp :ae STORE) ; not dynamic space
+          ;; sizeof (struct page) depends on GENCGC-CARD-BYTES
+          ;; It's 4+2+1+1 = 8 bytes if GENCGC-CARD-BYTES is (unsigned-byte 16),
+          ;; or   4+4+1+1 = 10 bytes (rounded to 12) if wider than (unsigned-byte 16).
+          ;; See the corresponding alien structure definition in 'room.lisp'
+          (cond ((typep gencgc-card-bytes '(unsigned-byte 16))
+                 (inst shl eax-tn 3) ; multiply by 8
+                 (inst add eax-tn (thread-slot-ea thread-dynspace-pte-base-slot)
+                       prefix)
+                 ;; clear WP - bit index 5 of flags byte
+                 (inst and (make-ea :byte :base eax-tn :disp 6) (lognot (ash 1 5))
+                       :lock))
+                (t
+                 (inst lea eax-tn ; multiply by 3
+                       (make-ea :dword :base eax-tn :index eax-tn :scale 2))
+                 (inst shl eax-tn 2) ; then by 4, = 12
+                 (inst add eax-tn (thread-slot-ea thread-dynspace-pte-base-slot)
+                       prefix)
+                 ;; clear WP
+                 (inst and (make-ea :byte :base eax-tn :disp 8) (lognot (ash 1 5))
+                       :lock)))
+          STORE
+          (inst mov edi-tn object)
+          (inst mov edx-tn word-index)
+          (inst mov eax-tn newval)
+          ;; set 'written' flag in the code header
+          (inst or (make-ea :byte :base edi-tn :disp (- 3 other-pointer-lowtag))
+                #x40 :lock)
+          ;; store newval into object
+          (inst mov (make-ea :dword :base edi-tn
+                             :index edx-tn :scale (ash 1 word-shift)
+                             :disp (- other-pointer-lowtag))
+                eax-tn)))))
+  (inst pop edi-tn) ; restore
+  (inst pop edx-tn)
+  (inst pop eax-tn)
+  (inst ret 12)) ; remove 3 stack args
