@@ -994,21 +994,6 @@ and no value was provided for it." name))))))))))
                    unknown-keys)
           (funcall unknown-keys-fun))))))
 
-(defun assert-modifying-lvar-type (lvar type caller policy)
-  (let ((internal-lvar (make-lvar))
-        (dest (lvar-dest lvar)))
-    (substitute-lvar internal-lvar lvar)
-    (with-ir1-environment-from-node dest
-      (let ((cast (make-modifying-cast
-                   :asserted-type type
-                   :type-to-check (maybe-weaken-check type policy)
-                   :value lvar
-                   :derived-type (coerce-to-values type)
-                   :caller caller)))
-        (%insert-cast-before dest cast)
-        (use-lvar cast internal-lvar)
-        t))))
-
 (defun assert-array-index-lvar-type (lvar type policy)
   (let ((internal-lvar (make-lvar))
         (dest (lvar-dest lvar)))
@@ -1023,38 +1008,30 @@ and no value was provided for it." name))))))))))
         (use-lvar cast internal-lvar)
         t))))
 
-(defun assert-cast-with-hook (lvar type policy
-                              hook)
-  (let ((internal-lvar (make-lvar))
-        (dest (lvar-dest lvar)))
-    (substitute-lvar internal-lvar lvar)
-    (with-ir1-environment-from-node dest
-      (let ((cast (make-cast-with-hook
-                   :asserted-type type
-                   :type-to-check (maybe-weaken-check type policy)
-                   :value lvar
-                   :derived-type (coerce-to-values type)
-                   :hook hook)))
-        (%insert-cast-before dest cast)
-        (use-lvar cast internal-lvar)
-        t))))
-
 (defun apply-type-annotation (fun-name arg type lvars policy &optional annotation)
   (case (car annotation)
     (function-designator
      (assert-function-designator fun-name lvars arg (cdr annotation))
      t)
-    (modifying
-     (if (policy policy (> check-constant-modification 0))
-         (assert-modifying-lvar-type arg type fun-name policy)
-         (assert-lvar-type arg type policy)))
-    (type-specifier
-     (assert-cast-with-hook arg type policy
-                            (lambda (value)
-                              (unless (and (eql value :default)
-                                           (eq fun-name 'open))
-                                (compiler-specifier-type value)))))
     (t
+     (case (car annotation)
+       (modifying
+        (when (policy policy (> check-constant-modification 0))
+          (add-annotation arg
+                          (make-lvar-modified-annotation :caller fun-name))))
+       (type-specifier
+        (add-annotation arg
+                        (make-lvar-type-spec-annotation
+                         :hook
+                         (lambda (value)
+                           (unless (and (eql value :default)
+                                        (eq fun-name 'open))
+                             (compiler-specifier-type value))))))
+       ((proper-list proper-sequence
+                     proper-or-circular-list proper-or-dotted-list)
+        (add-annotation arg
+                        (make-lvar-proper-sequence-annotation
+                         :kind (car annotation)))))
      (assert-lvar-type arg type policy))))
 
 ;;; Assert that CALL is to a function of the specified TYPE. It is
@@ -1226,3 +1203,16 @@ and no value was provided for it." name))))))))))
                                  (lvar-value detail)
                                  :cast-context (lvar-value cast-context))
   (ir2-convert-full-call node block))
+
+(defoptimizer (%compile-time-type-style-warn ir2-convert)
+    ((objects atype dtype detail code-context cast-context) node block)
+  (declare (ignore objects code-context block))
+  ;; Remove %COMPILE-TIME-TYPE-ERROR bits
+  (setf (node-source-path node)
+        (cddr (node-source-path node)))
+  (%compile-time-type-error-warn node
+                                 (lvar-value atype)
+                                 (lvar-value dtype)
+                                 (lvar-value detail)
+                                 :cast-context (lvar-value cast-context)
+                                 :condition 'type-style-warning))
