@@ -20,6 +20,7 @@
 
 ;;; Bind this to a stream to capture various internal debugging output.
 (defvar *compiler-trace-output* nil)
+(defvar *compile-trace-targets* '(:ir1 :ir2 :vop :disassemble))
 
 ;;; The current block compilation state. These are initialized to the
 ;;; :BLOCK-COMPILE and :ENTRY-POINTS arguments that COMPILE-FILE was
@@ -100,11 +101,6 @@
 ;; Used during compilation to keep track of with source paths have been
 ;; instrumented in which blocks.
 (defun code-coverage-blocks (x) (cdr x))
-;; Stores the code coverage instrumentation results. Keys are namestrings, the
-;; value is a list of (CONS PATH STATE), where STATE is +CODE-COVERAGE-UNMARKED+
-;; for a path that has not been visited, and T for one that has.
-(defvar *code-coverage-info* (make-hash-table :test 'equal))
-
 
 ;;;; WITH-COMPILATION-UNIT and WITH-COMPILATION-VALUES
 
@@ -373,7 +369,6 @@ Examples:
                      `(make-finite-sb
                        :conflicts (make-array ,size :initial-element #())
                        :always-live (make-array ,size :initial-element #*)
-                       :always-live-count (make-array ,size :initial-element 0)
                        :live-tns (make-array ,size :initial-element nil)))))))
      (unwind-protect
          (let ((*warnings-p* nil)
@@ -529,6 +524,7 @@ necessary, since type inference may take arbitrarily long to converge.")
         (maybe-mumble "constraint ")
         (constraint-propagate component))
       (when (retry-delayed-ir1-transforms :constraint)
+        (setf loop-count 0) ;; otherwise nothing may get retried
         (maybe-mumble "Rtran "))
       (flet ((want-reoptimization-p ()
                (or (component-reoptimize component)
@@ -553,11 +549,14 @@ necessary, since type inference may take arbitrarily long to converge.")
   (ir1-finalize component)
   (values))
 
+#!-immobile-code
+(defun component-mem-space (component)
+  (component-%mem-space component))
+
 #!+immobile-code
 (progn
-  (declaim (type (member :immobile :dynamic)
-                 *compile-to-memory-space*
-                 *compile-file-to-memory-space*))
+  (declaim (type (member :immobile :dynamic :auto) *compile-to-memory-space*)
+           (type (member :immobile :dynamic) *compile-file-to-memory-space*))
   ;; COMPILE-FILE puts all nontoplevel code in immobile space, but COMPILE
   ;; offers a choice. Because the collector does not run often enough (yet),
   ;; COMPILE usually places code in the dynamic space managed by our copying GC.
@@ -565,15 +564,22 @@ necessary, since type inference may take arbitrarily long to converge.")
   ;; The real default is set to :DYNAMIC in make-target-2-load.lisp
   (defvar *compile-to-memory-space* :immobile) ; BUILD-TIME default
   (defvar *compile-file-to-memory-space* :immobile) ; BUILD-TIME default
+  (defun component-mem-space (component)
+    (or (component-%mem-space component)
+        (setf (component-%mem-space component)
+              (if (fasl-output-p *compile-object*)
+                  (and (eq *compile-file-to-memory-space* :immobile)
+                       (neq (component-kind component) :toplevel)
+                       :immobile)
+                      *compile-to-memory-space*))))
   (defun code-immobile-p (thing)
-    (if (fasl-output-p *compile-object*)
-        (and (eq *compile-file-to-memory-space* :immobile)
-             (neq (component-kind (typecase thing
-                                    (vop  (node-component (vop-node thing)))
-                                    (node (node-component thing))
-                                    (t    thing)))
-                  :toplevel))
-        (eq *compile-to-memory-space* :immobile))))
+    #+sb-xc-host (declare (ignore thing)) #+sb-xc-host t
+    #-sb-xc-host
+    (let ((component (typecase thing
+                       (vop  (node-component (vop-node thing)))
+                       (node (node-component thing))
+                       (t    thing))))
+      (eq (component-mem-space component) :immobile))))
 
 (defun %compile-component (component)
   (let ((*adjustable-vectors* nil)) ; Needed both by codegen and fasl writer
@@ -641,32 +647,46 @@ necessary, since type inference may take arbitrarily long to converge.")
           (ir2-optimize-jumps component)
           (optimize-constant-loads component)
           (when *compiler-trace-output*
-            (describe-component component *compiler-trace-output*)
-            (describe-ir2-component component *compiler-trace-output*))
+            (when (memq :ir1 *compile-trace-targets*)
+              (describe-component component *compiler-trace-output*))
+            (when (memq :ir2 *compile-trace-targets*)
+             (describe-ir2-component component *compiler-trace-output*)))
 
           (maybe-mumble "code ")
-
-          (multiple-value-bind (segment code-length elsewhere-label fixup-notes)
-              (let (#!+immobile-code
-                    (*code-is-immobile* (code-immobile-p component)))
+          (multiple-value-bind (segment text-length fun-table
+                                elsewhere-label fixup-notes)
+              (let ((*compiler-trace-output*
+                      (and (memq :vop *compile-trace-targets*)
+                           *compiler-trace-output*)))
                 (generate-code component))
 
-            #-sb-xc-host
-            (when *compiler-trace-output*
-              (format *compiler-trace-output*
-                      "~|~%disassembly of code for ~S~2%" component)
-              (sb!disassem:disassemble-assem-segment segment
-                                                     *compiler-trace-output*))
-
-            (let ((object *compile-object*)
+            (let ((bytes (sb!assem:segment-contents-as-vector segment))
+                  (object *compile-object*)
                   (*elsewhere-label* elsewhere-label)) ; KLUDGE
+
+              (when (and *compiler-trace-output*
+                         (memq :disassemble *compile-trace-targets*))
+                (let ((ranges
+                        (maplist (lambda (list)
+                                   (cons (+ (car list)
+                                            (ash sb!vm:simple-fun-code-offset
+                                                 sb!vm:word-shift))
+                                         (or (cadr list) text-length)))
+                                 fun-table)))
+                  (declare (ignorable ranges))
+                  (format *compiler-trace-output*
+                          "~|~%disassembly of code for ~S~2%" component)
+                  #-sb-xc-host
+                  (sb!disassem:disassemble-assem-segment
+                   bytes ranges *compiler-trace-output*)))
+
               (funcall (etypecase object
-                        (fasl-output (maybe-mumble "fasl") #'fasl-dump-component)
-                        #-sb-xc-host ; no compiling to core
-                        (core-object (maybe-mumble "core") #'make-core-component)
-                        (null (lambda (&rest dummies)
-                                (declare (ignore dummies)))))
-                       component segment code-length fixup-notes
+                         (fasl-output (maybe-mumble "fasl") #'fasl-dump-component)
+                         #-sb-xc-host   ; no compiling to core
+                         (core-object (maybe-mumble "core") #'make-core-component)
+                         (null (lambda (&rest dummies)
+                                 (declare (ignore dummies)))))
+                       component segment (length bytes) fixup-notes
                        object))))))
 
   ;; We're done, so don't bother keeping anything around.
@@ -1635,7 +1655,9 @@ necessary, since type inference may take arbitrarily long to converge.")
          (let ((ctxt *compiler-error-context*))
            (lexenv-handled-conditions
             (etypecase ctxt
-             (node (node-lexenv ctxt))
+              (node (node-lexenv ctxt))
+              (lvar-annotation
+               (lvar-annotation-lexenv ctxt))
              (compiler-error-context
               (let ((lexenv (compiler-error-context-lexenv ctxt)))
                 (aver lexenv)
@@ -1724,7 +1746,7 @@ necessary, since type inference may take arbitrarily long to converge.")
                   ;; Dump the code coverage records into the fasl.
                    (with-source-paths
                     (fopcompile `(record-code-coverage
-                                  ',(namestring *compile-file-pathname*)
+                                  ',(namestring sb!xc:*compile-file-pathname*)
                                   ',(let (list)
                                       (maphash (lambda (k v)
                                                  (declare (ignore k))
