@@ -94,9 +94,38 @@
   (dynamic-extent nil :type (or null cleanup))
   ;; something or other that the back end annotates this lvar with
   (info nil)
-  (dependent-casts nil))
+  (dependent-casts nil)
+  (annotations nil))
 (!set-load-form-method lvar (:xc :target) :ignore-it)
 
+;;; These are used for annottating a LVAR with information that can't
+;;; be expressed using types.
+;;; Right now it's basically used for tracking constants and checking
+;;; them for things like proper sequence, or valid type specifier.
+(defstruct lvar-annotation
+  (source-path nil :type list)
+  lexenv
+  fired)
+
+(defprinter (lvar-annotation)
+  fired)
+
+(defstruct (lvar-modified-annotation
+            (:include lvar-annotation))
+  caller)
+
+(defstruct (lvar-hook
+            (:include lvar-annotation))
+  (hook #'missing-arg :type function))
+
+(defstruct (lvar-type-spec-annotation
+            (:include lvar-hook)))
+
+(defstruct (lvar-proper-sequence-annotation
+            (:include lvar-annotation))
+  (kind :sequence :type (member proper-list proper-sequence
+                                proper-or-circular-list proper-or-dotted-list)))
+
 (defmethod print-object ((x lvar) stream)
   (print-unreadable-object (x stream :type t :identity t)
     (when (boundp '*compiler-ir-obj-map*)
@@ -394,6 +423,10 @@
                          (outer-loop (make-loop :kind :outer :head head)))))
   ;; unique ID for debugging
   #!+sb-show (id (new-object-id) :read-only t)
+  ;; space where this component will be allocated in
+  ;; :auto won't make any codegen optimizations pertinent to immobile space,
+  ;; but will place the code there given sufficient available space.
+  (%mem-space nil :type (member nil :dynamic :immobile :auto))
   ;; the kind of component
   ;;
   ;; (The terminology here is left over from before
@@ -1291,8 +1324,6 @@
   (eql-var-constraints     nil :type (or null (array t 1)))
   (inheritable-constraints nil :type (or null (array t 1)))
   (private-constraints     nil :type (or null (array t 1)))
-  ;; Initial type of a LET variable as last seen by PROPAGATE-FROM-SETS.
-  (last-initial-type *universal-type* :type ctype)
   ;; The FOP handle of the lexical variable represented by LAMBDA-VAR
   ;; in the fopcompiler.
   (fop-value nil))
@@ -1333,7 +1364,9 @@
   ;; CONSTANT nodes are always anonymous, since we wish to coalesce named and
   ;; unnamed constants that are equivalent, we need to keep track of the
   ;; reference name for XREF.
-  (%source-name (missing-arg) :type symbol :read-only t))
+  (%source-name (missing-arg) :type symbol :read-only t)
+  ;; Constraints that cannot be expressed as NODE-DERIVED-TYPE
+  constraints)
 (defprinter (ref :identity t)
   #!+sb-show id
   (%source-name :test (neq %source-name '.anonymous.))
@@ -1489,7 +1522,7 @@
   (context nil)
   ;; Avoid compile time type conflict warnings.
   ;; Used by things that expand into ETYPECASE.
-  (silent-conflict nil :type boolean))
+  (silent-conflict nil :type (or boolean (eql :style-warning))))
 (defprinter (cast :identity t)
   %type-check
   value
@@ -1537,14 +1570,8 @@
   (arg-specs :test arg-specs)
   (result-specs :test result-specs))
 
-;;; Something will modify this value, warn if it's a constant
-(def!struct (modifying-cast (:include cast) (:copier nil))
-  ;; NIL after a warning has been issued
-  caller)
-
-;;; The hook is called when the value becomes constant
-(def!struct (cast-with-hook (:include cast) (:copier nil))
-  (hook #'missing-arg :type (or null function)))
+;;; Inserted by ARRAY-CALL-TYPE-DERIVER so that it can be later deleted
+(def!struct (array-index-cast (:include cast) (:copier nil)))
 
 ;;;; non-local exit support
 ;;;;
